from __future__ import annotations

from typing import Any, Callable

import os
import signal
from functools import partial

import psutil
from psutil import Process

_mem_unit_table = {
    "B": 1,
    "KB": 2**10,
    "MB": 2**20,
    "GB": 2**30,
}

_time_unit_table = {"s": 1, "m": 60, "h": 60 * 60}


def memconvert(x: float, *, frm: str = "B", to: str = "B") -> float:
    """Convert between memory units

    Parameters
    ----------
    x : float
        The memory amount

    frm : "B" | "KB" | "MB" | "GB" = "B"
        What unit it is in

    to :  "B" | "KB" | "MB" | "GB" = "B"
        What unit to convert to

    Returns
    -------
    float
        The memory amount
    """
    u_from = _mem_unit_table[frm.upper()]
    u_to = _mem_unit_table[to.upper()]

    as_bytes = x * u_from
    as_target = as_bytes / u_to

    # We can't see a use case for float Bytes
    if to.upper() == "B":
        return round(as_target)
    else:
        return as_target


def timeconvert(x: float, *, frm: str = "s", to: str = "s") -> float:
    """Convert between time units

    Parameters
    ----------
    x: float
        The time amount

    frm: "s" | "m" | "h" = "s"
        What unit it is in

    to: "s" | "m" | "h" = "s"
        What unit to convert to

    Returns
    -------
    float
        The time amount
    """
    u_from = _time_unit_table[frm.lower()]
    u_to = _time_unit_table[to.lower()]

    as_seconds = x * u_from
    as_target = as_seconds / u_to

    return as_target


def callstring(f: Callable, *args: Any, **kwargs: Any) -> str:
    """Get a string of the function being called with the args and kwargs

    Parameters
    ----------
    f: Callable
        The function

    *args, **kwargs

    Returns
    -------
    str
        The function call as a str
    """
    parts = list(map(str, args)) + [f"{k}={v}" for k, v in kwargs.items()]
    param_str = ", ".join(parts)
    if isinstance(f, partial):
        name = f.func.__name__
    elif hasattr(f, "__qualname__"):
        name = f.__qualname__
    elif hasattr(f, "__class__"):
        name = f.__class__.__name__
    else:
        name = str(f)

    return f"{name}({param_str})"


def terminate_process(
    pid: int | None | psutil.Process = None,
    sig: int = signal.SIGTERM,
    timeout: int = 5,
    children: bool = True,
    parent: bool = False,
    on_terminate: Callable[[psutil.Process], Any] | None = None,
) -> tuple[list[psutil.Process], list[psutil.Process]]:
    """Attemps to terminate a process

    * https://psutil.readthedocs.io/en/latest/#kill-process-tree
    """
    if children is True:
        return terminate_process_tree(
            pid=pid,
            include_parent=parent,
            sig=sig,
            timeout=timeout,
        )
    else:
        if isinstance(pid, psutil.Process):
            process = pid
        else:
            try:
                process = psutil.Process(pid)
            except psutil.NoSuchProcess:
                return ([], [])

        process.send_signal(sig)
        gone, alive = psutil.wait_procs(
            [process], timeout=timeout, callback=on_terminate
        )

        for p in alive:
            p.kill()

        gone, alive = psutil.wait_procs(
            [process], timeout=timeout, callback=on_terminate
        )
        return (gone, alive)


def terminate_process_tree(
    pid: int | None | psutil.Process = None,
    sig: int = signal.SIGTERM,
    timeout: int = 5,
    include_parent: bool = True,
    on_terminate: Callable[[psutil.Process], Any] | None = None,
) -> tuple[list[psutil.Process], list[psutil.Process]]:
    """Attemps to get all the children of the process and then terminate them

    * https://psutil.readthedocs.io/en/latest/#kill-process-tree
    """
    parent = None

    if isinstance(pid, psutil.Process):
        parent = pid
        _pid = parent.pid
    elif isinstance(pid, int):
        _pid = pid
    else:
        _pid = os.getpid()

    if _pid == os.getpid() and include_parent:
        raise RuntimeError(f"Can't kill this process ({pid}) from within itself")

    # Get the parent and its children
    try:
        if parent is None:
            parent = psutil.Process(pid=_pid)

        children = parent.children(recursive=True)
    except psutil.NoSuchProcess:
        children = []

    if include_parent and parent is not None:
        children.append(parent)

    for child in children:
        try:
            child.send_signal(sig)
        except psutil.NoSuchProcess:
            pass

    gone, alive = psutil.wait_procs(children, timeout=timeout, callback=on_terminate)

    # If any are still alive, SIGKILL
    for child in alive:
        try:
            child.kill()
        except psutil.NoSuchProcess:
            pass

    gone, alive = psutil.wait_procs(children, timeout=None, callback=on_terminate)
    return (gone, alive)


class Monitor:
    def __init__(self, pid: int | None = None):
        """
        Parameters
        ----------
        pid : int | None = None
            The process id to monitor, defaults to current process
        """
        self.process = Process(pid)

    def memory(self, units: str = "B", *, kind: str = "vms") -> float:
        """Get the memory consumption

        Parameters
        ----------
        units : "B" | "KB" | "MB" | "GB" = "B"
            Units to measure in

        kind : "vms" | "rss" = "vms"
            The kind of memory to measure.
            https://psutil.readthedocs.io/en/latest/#psutil.Process.memory_info

        Returns
        -------
        float
            The memory used
        """
        mem = self.process.memory_info()
        if not hasattr(mem, kind):
            raise ValueError(f"No memory kind {kind}, use one from {mem}")

        usage = getattr(mem, kind)
        return memconvert(usage, frm="B", to=units)

    def memlimit(self, units: str = "B") -> tuple[float, float] | None:
        """

        We can't limit using resource.setrlimit as it seems that None of the
        RLIMIT_X's are available. This we debugged by using
        `import psutil; print(dir(psutil))` in which a MAC system did not have
        any `RLIMIT_X` attributes while a Linux system did.

        Parameters
        ----------
        units : "B" | "KB" | "MB" | "GB" = "B"
            Units to measure in

        Returns
        -------
        float | None
            The memory limit.
            Returns None if it can't be gotten
        """
        if hasattr(psutil, "RLIMIT_AS"):
            limits = self.process.rlimit(psutil.RLIMIT_AS)
            if units != "B":
                limits = tuple(memconvert(x, frm="B", to=units) for x in limits)
            return limits
        else:
            return None



from pynisher.exceptions import (
    CpuTimeoutException,
    MemoryLimitException,
    PynisherException,
    TimeoutException,
    WallTimeoutException,
)
from pynisher.pynisher import EMPTY, Pynisher, limit, restricted
from pynisher.support import (
    contexts,
    supports,
    supports_cputime,
    supports_limit_decorator,
    supports_memory,
    supports_walltime,
)

__all__ = [
    "Pynisher",
    "EMPTY",
    "limit",
    "restricted",
    "CpuTimeoutException",
    "MemoryLimitException",
    "PynisherException",
    "TimeoutException",
    "WallTimeoutException",
    "contexts",
    "supports",
    "supports_walltime",
    "supports_cputime",
    "supports_memory",
    "supports_limit_decorator",
]



"""
Useful info for invesetigating memory:
* MacOS uses Darwin-XNU as it's underlying operating system. This is easier to find as
just XNU.
* XNU uses a special memory manager called Mach-VM in its interface to which we have no
control
* Darwin-xnu: https://github.com/apple/darwin-xnu
* resource.h: https://github.com/apple/darwin-xnu/blob/main/bsd/sys/resource.h
* kern_resource.c
    https://github.com/apple/darwin-xnu/blob/main/bsd/kern/kern_resource.c

For documentation on `setrlimit` and how to limit resources for MAC
** `setrlimit` **
* https://github.com/apple/darwin-xnu/blob/main/bsd/man/man2/getrlimit.2

For documentation on MAC specific signal alarms:
** `signals` **
https://developer.apple.com/library/archive/documentation/System/Conceptual/ManPages_iPhoneOS/man2/sigaction.2.html

Can't seem to limit memory, the only accepted rlimit towards memory is RLIMIT_DATA and
that seems to not work.
# https://github.com/apple/darwin-xnu/blob/2ff845c2e033bd0ff64b5b6aa6063a1f8f65aa32/bsd/kern/kern_resource.c#L987

This module and the Linux limiter share almost identical code but we keep them seperate
incase of specific modules or changes needed
"""
from __future__ import annotations

import resource

from pynisher.limiters.limiter import Limiter


class LimiterMac(Limiter):
    def limit_memory(self, memory: int) -> None:
        """Limit the addressable memory

        It seems that each of RLIMIT_AS, RLIMIT_DATA and RLIMIT_RSS do nothing.
        While they do set, nothing is done when those boundaries are exceeded.

        * Tried catching all available signals but non triggered

        We still however try this but it's unlikely to work

        Parameters
        ----------
        memory : int
            The memory limit in bytes
        """
        # This will likely do nothing on newer mac, however users can check for support
        # before hand to prevent issues.
        try:
            soft, hard = resource.getrlimit(resource.RLIMIT_AS)

            self.old_limits = (soft, hard)
            new_limits = (memory, hard)

            resource.setrlimit(resource.RLIMIT_AS, new_limits)
        except Exception:
            raise RuntimeError("Limiting memory is not supported on your platform.")

    def limit_cpu_time(self, cpu_time: int, interval: int = 5) -> None:
        """Limit the cpu time for this process.

        The process will be killed with -signal.SIGXCPU status.

        Attempts to handle the sigxcpu with `signal.signal(signal.SIGXCPU, handler)`
        did not work as intended in cases where the process spawn subprocesses

        Using `setitimer(ITIMER_PROF)` is not a good idea as it only measure the time
        of the current process, non of it's children.


        Parameters
        ----------
        cpu_time : int
            The amount of time in seconds

        interval: int = 5
            The time between the initial SIGXCPU sent before a SIGKILL will be
            sent if the process is still running. This SIGXCPU is sent every
            second during this interval. See documentation linked in module
        """
        limit = (cpu_time, cpu_time + interval)
        resource.setrlimit(resource.RLIMIT_CPU, limit)



"""Set limits for Linux.

For documentation on `setrlimit` and how to limit resources for Linux
** `setrlimit` **
https://man7.org/linux/man-pages/man2/setrlimit.2.html

For documentation on Linux specific signal alarms:
** `signals` **
https://man7.org/linux/man-pages/man7/signal.7.html

This module and the Mac limiter share almost identical code but we keep them seperate
incase of specific modules or changes needed.
"""
from __future__ import annotations

import resource

from pynisher.limiters.limiter import Limiter


class LimiterLinux(Limiter):
    def limit_memory(self, memory: int) -> None:
        """Limit the addressable memory.

        This could technically raise `SIGSEGV` (segmentation fault) but
        we instead catch a python `MemoryError` as indication that memory
        time was exceeded. This lets us give back the traceback.

        Parameters
        ----------
        memory : int
            The memory limit in bytes
        """
        soft, hard = resource.getrlimit(resource.RLIMIT_AS)

        self.old_limits = (soft, hard)
        new_limits = (memory, hard)

        resource.setrlimit(resource.RLIMIT_AS, new_limits)

    def limit_cpu_time(self, cpu_time: int, interval: int = 5) -> None:
        """Limit the cpu time for this process.

        The process will be killed with -signal.SIGXCPU status

        Attempts to handle the sigxcpu with `signal.signal(signal.SIGXCPU, handler)`
        did not work as intended in cases where the process spawn subprocesses

        Using `setitimer(ITIMER_PROF)` is not a good idea as it only measure the time
        of the current process, non of it's children.

        Parameters
        ----------
        cpu_time : int
            The amount of time in seconds

        interval: int = 5
            The time between the initial SIGXCPU sent before a SIGKILL will be
            sent if the process is still running. This SIGXCPU is sent every
            second during this interval. See documentation linked in module
        """
        limit = (cpu_time, cpu_time + interval)
        resource.setrlimit(resource.RLIMIT_CPU, limit)

    def _debug_memory(self) -> str:
        """Prints the memory used by the process

        Returns
        -------
        (usage: str)
            Returns the usage as a string, not sure if always KB but
            leaving type as str until known.
        """
        # https://stackoverflow.com/a/39765583/5332072
        with open("/proc/self/status") as f:
            status = f.readlines()

        vmpeak = next(s for s in status if s.startswith("VmPeak:"))
        return vmpeak



from __future__ import annotations

from typing import Any, Callable, Generic, Type, TypeVar, overload

import multiprocessing
import signal
import sys
from functools import wraps

import psutil
from typing_extensions import Literal, ParamSpec

from pynisher.exceptions import (
    CpuTimeoutException,
    MemoryLimitException,
    PynisherException,
    TimeoutException,
    WallTimeoutException,
)
from pynisher.limiters import Limiter
from pynisher.support import contexts as valid_contexts
from pynisher.support import supports
from pynisher.util import callstring, memconvert, terminate_process, timeconvert
from pynisher.win_errcodes import WIN_CPUTIMEOUT_EXITCODES, WIN_MEMORY_EXITCODES


class _EMPTY:
    pass


EMPTY = _EMPTY()

# For typing the pynished function
T = TypeVar("T")
P = ParamSpec("P")

# For returning the same type as itself, see __enter__
Self = TypeVar("Self")


class Pynisher(Generic[P, T]):
    """Restrict a function's resources"""

    WallTimeoutException = WallTimeoutException
    CpuTimeoutException = CpuTimeoutException
    MemoryLimitException = MemoryLimitException
    PynisherException = PynisherException
    TimeoutException = TimeoutException

    # If `raises=True` or left as default, the return type when calling is just T
    @overload
    def __init__(
        self: Pynisher[P, T],
        func: Callable[P, T],
        *,
        raises: Literal[True] = ...,
        name: str | None = ...,
        memory: int | tuple[int, str] | None = ...,
        cpu_time: int | tuple[float, str] | None = ...,
        wall_time: int | tuple[float, str] | None = ...,
        context: str | None = ...,
        warnings: bool = ...,
        wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = ...,
        terminate_child_processes: bool = ...,
        forceful_keyboard_interrupt: bool = ...,
    ) -> None:
        ...

    # If `raises=False` or just some unknown bool value,
    # the return type when calling is T | _EMPTY
    @overload
    def __init__(
        self: Pynisher[P, T | _EMPTY],
        func: Callable[P, T],
        *,
        raises: bool,
        name: str | None = ...,
        memory: int | tuple[int, str] | None = ...,
        cpu_time: int | tuple[float, str] | None = ...,
        wall_time: int | tuple[float, str] | None = ...,
        context: str | None = ...,
        warnings: bool = ...,
        wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = ...,
        terminate_child_processes: bool = ...,
        forceful_keyboard_interrupt: bool = ...,
    ) -> None:
        ...

    def __init__(
        self,
        func: Callable[P, T],
        *,
        name: str | None = None,
        memory: int | tuple[int, str] | None = None,
        cpu_time: int | tuple[float, str] | None = None,
        wall_time: int | tuple[float, str] | None = None,
        context: str | None = None,
        raises: bool = True,
        warnings: bool = True,
        wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = False,
        terminate_child_processes: bool = True,
        forceful_keyboard_interrupt: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        func : Callable
            The function to limit and call

        name : str | None
            A name to give the process that gets created, defaults to whatever
            multiprocessing.Process defaults to.

        memory : int | tuple[int, str] | None = None
            The amount of memory to limit by. If `tuple`, specify units with (4, "MB").
            Possible units are "B", "KB", "MB", "GB".

            Processes take up some space before any limitation can take place.
            These will run fine until a new allocation takes place.
            This means a process can technically run in a limit of 1 Byte, up until the
            point it tries to allocate anything.

        cpu_time : int | tuple[float, str] | None = None
            The cpu time in seconds to limit the process to. This time is only counted while the
            process is active.
            Can provide in (time, units) such as (1.5, "h") to indicate one and a half hours.
            Units available are "s", "m", "h".

        wall_time : int | tuple[float, str] | None = None
            The amount of total wall time in seconds to limit to
            Can provide in (time, units) such as (1.5, "h") to indicate one and a half hours.
            Units available are "s", "m", "h"

        context : "fork" | "forkserver" | "spawn" | None = None
            The context to use with multiprocessing.get_context()
            * https://docs.python.org/3/library/multiprocessing.html#multiprocessing.get_context

        raises : bool = True
            Whether any error from the subprocess should filter up and be raised.

        warnings : bool = True
            Whether to emit pynisher warnings or not.

        wrap_errors: bool | list[str | Type[Exception]] | dict = False
            Whether exceptions raised due to none resource related issues should be
            wrapped in a PynisherException.

            * If True, all exceptions will be wrapped

            * If False, no exceptions will be wrapped

            * If list, you can provide your own list of exceptions to wrap, using a
                string to prevent explicit imports outside the limited function.

            * If dict, you can specify a mapping to which kind of pynisher errors map
                to which kind of Exceptions. Again, you can specify a str if needed.
                Please see below.

            For advanced users, this can let you control whether module information will
            get sent back from the limited function.

            .. code:: python

                def f():
                    import sklearn
                    raise sklearn.exceptions.NotFittedError()

                lf = limit(f, wrap_errors=True):
                lf = limit(f, wrap_errors=["NotFittedError"])  # More specific

                try:
                    lf()
                except PynisherException as e:
                    # The error `e` has no context related to sklearn so it's not
                    # imported here

            The dict arg lets you indicated how particular errors should be identified.
            These will only be active if the corresponding limit is set.

            This can be useful in cases where a known error is caused due to memory
            constraints but does not raise a Python MemoryError.

            Keys: "cpu_time", "wall_time", "memory", "pynisher"

            Values: str, ExceptionType, (OSError, code), (OSError, code, winerr)

            .. code:: Python

                wrap_errors = {
                    "memory": [ImportError, (OSError, 22, 1455)]
                    "cpu_time": ["MyCustomException"]
                }

            For example, importing sklearn with a small enough memory limit can trigger
            an `ImportError` on linux or a `(OSError, 22, 1455)` on windows. Here the
            `22` stands for the Python OSError codes and the `1455` is the winerror code.

            In general, you should not have to interface with this but it can allow you
            some extra control.

        terminate_child_processes: bool = True
            Whether to clean up all child processes upon completion

        forceful_keyboard_interrupt: bool = True
            Whether keyboard interrupts should forceably kill any subprocess or the
            pynished function. If True, it will temrinate the process tree of
            the pynished function and then reraise the KeyboardInterrupt.
        """  # noqa
        _cpu_time: int | None
        if isinstance(cpu_time, tuple):
            x, unit = cpu_time
            _cpu_time = round(timeconvert(x, frm=unit))
        else:
            _cpu_time = cpu_time

        _wall_time: int | None
        if isinstance(wall_time, tuple):
            x, unit = wall_time
            _wall_time = round(timeconvert(x, frm=unit))
        else:
            _wall_time = wall_time

        _memory: int | None
        if isinstance(memory, tuple):
            x, unit = memory
            _memory = round(memconvert(x, frm=unit))
        else:
            _memory = memory

        if not callable(func):
            raise ValueError(f"`func` ({func}) must be callable")

        if _cpu_time is not None and not _cpu_time >= 1:
            raise ValueError(f"`cpu_time` {cpu_time} must be >= 1 seconds")

        if _wall_time is not None and not _wall_time >= 1:
            raise ValueError(f"`wall_time` {wall_time} must be >= 1 second")

        if _memory is not None and not _memory >= 1:
            raise ValueError(f"`memory` {memory} must be >= 1 Byte")

        if isinstance(context, str) and context not in valid_contexts:
            raise ValueError(f"`context` {context} must be in {valid_contexts}")

        if isinstance(wrap_errors, dict):
            valid_keys = {"memory", "wall_time", "cpu_time", "pynisher"}
            keys = list(wrap_errors.keys())
            if not all(key in valid_keys for key in keys):
                raise ValueError(
                    f"`wrap_errors` has unknown key in {keys},"
                    f" each must be in {valid_keys} "
                )

        self.func = func
        self.name = name
        self.cpu_time = _cpu_time
        self.memory = _memory
        self.wall_time = _wall_time
        self.raises = raises
        self.context = (
            multiprocessing.get_context(context)
            if isinstance(context, str) or context is None
            else context
        )
        self.warnings = warnings
        self.wrap_errors = wrap_errors
        self.terminate_child_processes = terminate_child_processes
        self.forceful_keyboard_interrupt = forceful_keyboard_interrupt

        # Set once the function is running
        self._process: psutil.Process | None = None

    def __enter__(self: Self) -> Self:
        """Doesn't do anything too useful at the moment.

        Returns
        -------
        (*args, **kwargs) -> Any
            Call your function and get back the result
        """
        return self

    def __exit__(self, *exc: Any) -> None:
        """Doesn't do anything too useful at the moment.

        Returns
        -------
        (*args, **kwargs) -> Any
            Call your function and get back the result
        """
        # Note: Not sure if we have to handle this
        # *https://docs.python.org/3/reference/datamodel.html#object.__exit__
        return

    # Call with `raises=True`
    @overload
    def __call__(
        self: Pynisher[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        ...

    # Call with `raises=False`
    @overload
    def __call__(
        self: Pynisher[P, T | _EMPTY],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T | _EMPTY:
        ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T | _EMPTY:
        """Run the function with the given set of arguments and block until result.

        Parameters
        ----------
        *args, **kwargs
            Parameters to pass to the function being limited

        Returns
        -------
        Any
            The result of calling the function that is being limited
        """
        # The send pipe will be passed to the subprocess to send the result back
        # while the receive pipe will be used to recieve it in this master process
        receive_pipe, send_pipe = self.context.Pipe(duplex=False)

        # The limiter is in charge of limiting resources once inside the subprocess
        # It gets the `send_pipe` through which it it should `output` it's results to
        limiter = Limiter.create(
            func=self.func,
            output=send_pipe,
            memory=self.memory,
            cpu_time=self.cpu_time,
            wall_time=self.wall_time,
            warnings=self.warnings,
            wrap_errors=self.wrap_errors,
            terminate_child_processes=self.terminate_child_processes,
        )

        # We now create the subprocess and let it know that it should call the limiter's
        # __call__ with the args and kwargs for the function being limited
        subprocess = self.context.Process(  # type: ignore
            target=limiter.__call__,
            args=args,  # type: ignore
            kwargs=kwargs,  # type: ignore
            daemon=False,
            name=self.name,
        )

        # Let loose
        subprocess.start()

        # Get a psutil handle to the process
        try:
            self._process = psutil.Process(subprocess.pid)
        except psutil.NoSuchProcess:
            # Likely only to occur when subprocess already finished
            pass

        # If self.wall time is None, block until the subprocess finishes or
        # terminates. Otherwise, will return after wall_time and the process
        # will still be running
        if not self.forceful_keyboard_interrupt:
            subprocess.join(self.wall_time)
        else:
            # The keyboard interrupt will be send to all processes simultaneuously
            # and handled by each of them. The default behaviour is to terminate
            # but this can be caught by a subprocess and ignored. To circumvent
            # this, we convert this keyboard interrupt into a SIGTERM and propgate it
            try:
                subprocess.join(self.wall_time)
            except KeyboardInterrupt:
                terminate_process(
                    subprocess.pid,
                    children=self.terminate_child_processes,
                    parent=True,
                )
                raise KeyboardInterrupt

        # exitcode here can only take on 3 values
        #
        # * None        | the subprocess is still running (walltime elapsed)
        # * 0           | Ended gracefully, pipe is closed or response in the pipe
        # * != 0        | Process was terminated non-gracefully, nothing in the pipe
        #               | and it may not be closed
        exitcode = subprocess.exitcode

        terminate_process(
            subprocess.pid,
            children=self.terminate_child_processes,
            parent=True,
        )

        # 4 kinds of `response` we expect
        #
        # * (result, None, None)    | success
        #                               => No error, we got a result
        # * (None, error, traceback)| failed
        #                               =>  Error, CPUtimeout linux/mac, MemoryError
        # * None                    | failed
        #                               => MemoryError during sending of real error
        # * EMPTY                   | failed, nothing received from pipe
        #                               => Walltime, CPUTimeout windows, Unknown
        result: _EMPTY | T = EMPTY
        err: Exception | None = None
        tb: str | None = None

        # Retrieve a result if we can,
        if receive_pipe.poll() and not receive_pipe.closed:
            try:
                response = receive_pipe.recv()
                if response is not None:
                    result, err, tb = response
                else:
                    result = EMPTY
                    tb = None
                    err = MemoryLimitException(
                        "While returning the result from your function,"
                        " we could not retrieve the result or any error"
                        " about why."
                        f"\n{callstring(self.func, *args, **kwargs)}",
                    )

            # Otherwise, there was nothing to read
            except EOFError:
                result = EMPTY
                tb = None
                err = MemoryLimitException(
                    "There but could not a send response back."
                    " This is likely a MemoryError."
                    f"\n{callstring(self.func, *args, **kwargs)}"
                )

        # Cleanup pipes
        receive_pipe.close()
        send_pipe.close()

        # We got a result or an error
        if result is not EMPTY or err is not None:
            # If an error, the result must be empty
            if err is not None:
                result = EMPTY

            return self._handle_return(result=result, err=err, tb=tb)

        # Process ended gracefully but no result?
        if exitcode == 0:
            err = PynisherException(
                f"Function ended properly but no result was recieved."
                f" Got exitcode 0 from subprocess that ran:"
                f"\n{callstring(self.func, *args, **kwargs)}"
            )
            return self._handle_return(err=err)

        # Wall time expired
        if exitcode is None:
            err = WallTimeoutException(
                f"Did not finish in time ({self.wall_time}s)"
                f"\n{callstring(self.func, *args, **kwargs)}"
            )
            return self._handle_return(err=err)

        # Cputime expired on windows
        if (
            sys.platform.lower().startswith("win")
            and self.cpu_time is not None
            and exitcode in WIN_CPUTIMEOUT_EXITCODES
        ):
            err = CpuTimeoutException(
                f"Did not finish in cpu time ({self.cpu_time}s)."
                f" Specific exitcode is {exitcode}"
                f"\n{callstring(self.func, *args, **kwargs)}"
            )
            return self._handle_return(err=err)

        # Memory reasons to kill process on windows
        if (
            sys.platform.lower().startswith("win")
            and self.memory is not None
            and exitcode in WIN_MEMORY_EXITCODES
        ):
            # We can't be certain it was caused by a cputimeout but the exist status
            # in this case is non-consisten and I do not know a way to identify the
            # object timed out properly
            err = MemoryLimitException(
                f"Not enough memory to run function ({self.memory}B)."
                f" Specific exitcode is {exitcode}"
                f"\n{callstring(self.func, *args, **kwargs)}"
            )
            return self._handle_return(err=err)

        # We got a segmentation fault, if memomory was set, we assume it's a memory
        # related issue
        if exitcode == -signal.SIGSEGV and self.memory is not None:
            err = MemoryLimitException(
                "The function exited with a segmentation error (SIGSEGV) and a memory"
                " limit was set. We presume this is due to the memory limit. This may"
                " not be the case but is quite likely if your function works without a"
                " memory constraint."
                f"\n{callstring(self.func, *args, **kwargs)}"
            )
            return self._handle_return(err=err)

        # We got a SIGXCPU and cputime was set
        if (
            self.cpu_time is not None
            and hasattr(signal, "SIGXCPU")
            and exitcode == -signal.SIGXCPU
        ):
            err = CpuTimeoutException(
                f"Did not finish in cpu time ({self.cpu_time}s)"
                f"\n{callstring(self.func, *args, **kwargs)}"
            )
            return self._handle_return(err=err)

        # We didn't get a result and the pipe closed in some way we weren't expecting
        err = PynisherException(
            f"Unknown reason for exitcode {exitcode}, no result or error recieved and "
            f" killed process \n{callstring(self.func, *args, **kwargs)}"
        )
        return self._handle_return(err=err)

    def _handle_return(
        self,
        result: T | _EMPTY = EMPTY,
        err: Exception | None = None,
        tb: str | None = None,
    ) -> T | _EMPTY:
        # We need at least an error or a result, not both
        assert (result is EMPTY) ^ (err is None)

        # We got a non empty result, hurray
        if result is not EMPTY:
            return result

        # Otherwise, we have some error
        assert err is not None

        # Don't raise?
        if not self.raises:
            return EMPTY

        if tb is not None:
            # Just so we can insert the traceback
            raise err from err.__class__(tb)
        else:
            raise err

    @staticmethod
    def supports(limit: str) -> bool:
        """Check if pynisher supports a given feature

        Parameters
        ----------
        limit: "wall_time" | "cpu_time" | "memory" | "decorator"
            The feature to check support for

        Returns
        -------
        bool
            Whether it is supported or not
        """
        return supports(limit)


@overload
def restricted(
    name: str | None = ...,
    *,
    raises: Literal[True] = ...,
    memory: int | tuple[int, str] | None = ...,
    cpu_time: int | None = ...,
    wall_time: int | None = ...,
    context: str | None = ...,
    warnings: bool = ...,
    wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = ...,
    terminate_child_processes: bool = ...,
    forceful_keyboard_interrupt: bool = True,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...


@overload
def restricted(
    name: str | None = ...,
    *,
    raises: bool,
    memory: int | tuple[int, str] | None = ...,
    cpu_time: int | None = ...,
    wall_time: int | None = ...,
    context: str | None = ...,
    warnings: bool = ...,
    wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = ...,
    terminate_child_processes: bool = ...,
    forceful_keyboard_interrupt: bool = True,
) -> Callable[[Callable[P, T]], Callable[P, T | _EMPTY]]:
    ...


# NOTE: Simpler solution?
#
#   There might be a simpler solution then redfining a function, e.g. `restricted =
#   Pynisher` but it gets complicated as we need something like
#   `@restricted(memory=...)` but that won't work as the first arg to
#   `Pynisher.__init__` should be the function itself. For now this should work
#
# Note: Having one positional argument
#
#   This just lets us catch the cases where someone incorrectly uses the decorator as
#   the first argument would be the function. Otherwise with only keyord arguments,
#   you would get a python exception that just says function doesn't except positinal
#   arguments
#
#   @restricted   # <- Should raise helpful error
#   def f(): ...
#
def restricted(
    name: str | None = None,
    *,
    memory: int | tuple[int, str] | None = None,
    cpu_time: int | None = None,
    wall_time: int | None = None,
    context: str | None = None,
    raises: bool = True,
    warnings: bool = True,
    wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = False,
    terminate_child_processes: bool = True,
    forceful_keyboard_interrupt: bool = True,
) -> Callable[[Callable[P, T]], Callable[P, T | _EMPTY]]:
    """Limit a function's resource consumption on each call

    ..code:: python

        @restricted(memory=1000, wall_time=14)
        def f(x: int) -> int:
            return x * 2

        f()

    Note
    ----
    Due to how multiprocessing pickling works, `@restricted(...)` does not
    work for Mac/Windows with Python >= 3.8. Please use the `Pynisher`"
    method of limiting resources in this case.

    Parameters
    ----------
    name : str | None
        A name to give the process that gets created, defaults to whatever
        multiprocessing.Process defaults to.

    memory : int | tuple[int, str] | None = None
        The amount of memory to limit by. If `tuple`, specify with units like (4, "MB").
        Possible units are "B", "KB", "MB", "GB".

        Processes are given some dedicated size before any limitation can take place.
        These will run fine until a new allocation takes place.
        This means a process can technically run in a limit of 1 Byte, up until the
        point it tries to allocate anything.

    cpu_time : int | tuple[float, str] | None = None
        The cpu time in seconds to limit the process to. This time is only counted while the
        process is active.
        Can provide in (time, units) such as (1.5, "h") to indicate one and a half hours.
        Units available are "s", "m", "h".

    wall_time : int | tuple[float, str] | None = None
        The amount of total wall time in seconds to limit to
        Can provide in (time, units) such as (1.5, "h") to indicate one and a half hours.
        Units available are "s", "m", "h"

    context : "fork" | "forkserver" | "spawn" | None = None
        The context to use with multiprocessing.get_context()
        * https://docs.python.org/3/library/multiprocessing.html#multiprocessing.get_context

    raises : bool = True
        Whether any error from the subprocess should filter up and be raised.

    warnings : bool = True
        Whether to emit pynisher warnings or not.

    wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = False
        Please see `Pynisher.__init__`

    terminate_child_processes: bool = True
        Whether to clean up all child processes upon completion

    forceful_keyboard_interrupt: bool = True
        Whether keyboard interrupts should forceably kill any subprocess or the
        pynished function. If True, it will temrinate the process tree of
        the pynished function and then reraise the KeyboardInterrupt.
    """  # noqa
    if not supports("decorator") or context == "spawn":
        raise ValueError(
            "Due to how multiprocessing pickling works, `@restricted(...)` does not"
            " for Mac or Windows, specifically with the `spawn` context."
            " Please use the `limit` method of limiting."
        )

    # Incase the first argument is a function, we assume it was missued
    #
    # @restricted
    # def f(): ...
    #
    # In this case, the function f will be passed as the first arg, `name`
    if callable(name):
        raise ValueError("Please pass arguments to decorator `@restricted`")

    def decorator(func: Callable[P, T]) -> Callable[P, T | _EMPTY]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T | _EMPTY:
            pynisher = Pynisher(
                func,
                name=name,
                memory=memory,
                cpu_time=cpu_time,
                wall_time=wall_time,
                raises=raises,
                wrap_errors=wrap_errors,
            )
            return pynisher(*args, **kwargs)

        return wrapper

    return decorator


limit = Pynisher



from pynisher.limiters.limiter import Limiter

# NOTE: no import of system limiters
#
#   Would likely cause cicrular imports and prevents any system
#   specific modules from being imported
__all__ = ["Limiter"]



# The master cheat-sheets
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/1bc92ddf-b79e-413c-bbaa-99a5281a6c90
#
# If you find out more codes caused by memory specific reasons for Windows to kill a
# process, please put in a PR to add them
# The cpu timeout one should be sufficient
#
# Do not import `pywin32` or other windows related libraries here as we need it to
# import on all systems

# Memory related
WIN_STATUS_PAGEFILE_QUOTA = 0xC0000007  # 3221225479
WIN_STATUS_IN_PAGE_ERROR = 0xC0000006  # 3221225478
WIN_STATUS_ACCESS_VIOLATION = 0xC0000005  # 3221225477

WIN_MEMORY_EXITCODES = [
    WIN_STATUS_PAGEFILE_QUOTA,
    WIN_STATUS_IN_PAGE_ERROR,
    WIN_STATUS_ACCESS_VIOLATION,
]

# Cputime out
WIN_STATUS_QUOTA_EXCEEDED = 0xC0000044  # 3221225540

WIN_CPUTIMEOUT_EXITCODES = [WIN_STATUS_QUOTA_EXCEEDED]

# Specific WinError codes attached ot an OSError
WIN_ERROR_COMMITMENT_LIMIT = 1455



class PynisherException(Exception):
    """Base class for any Pynisher related exceptions"""

    pass


class TimeoutException(PynisherException):
    """Base class for Timeout based errors"""

    pass


class CpuTimeoutException(TimeoutException):
    """Exception when hitting CPU time limit."""

    pass


class WallTimeoutException(TimeoutException):
    """Exception when hitting the wall time limit."""

    pass


class MemoryLimitException(PynisherException, MemoryError):
    """Exception when hitting the Memory Limit."""

    pass



"""Limiters are what define how resources should be limited as well as handle."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Type

import os
import platform
import signal
import sys
import traceback
from multiprocessing.connection import Connection

from pynisher.exceptions import (
    CpuTimeoutException,
    MemoryLimitException,
    PynisherException,
    WallTimeoutException,
)
from pynisher.util import Monitor, callstring, terminate_process
from pynisher.win_errcodes import WIN_ERROR_COMMITMENT_LIMIT


def is_err(err: Exception, err_type: str | Type[Exception]) -> bool:
    """Return if err is of a given type"""
    # Check name of class matches
    a = isinstance(err_type, str) and type(err).__name__ == err_type

    # Explicitly check for just class, not subclass
    # We do this because many library specific errors may inherit from base errors
    # and we don't want to accidentally wrap to many errors.
    # i.e. sklearn.exceptions.NotFittedError inherits form ValueError
    b = isinstance(err_type, type) and type(err) == err_type

    return a or b


class Limiter(ABC):
    """Defines how to limit resources for a given system."""

    def __init__(
        self,
        func: Callable,
        output: Connection,
        memory: int | None = None,
        cpu_time: int | None = None,
        wall_time: int | None = None,
        warnings: bool = True,
        wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = False,
        terminate_child_processes: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        func : Callable
            The function to be limited

        output : Connection
            The output multiprocessing.Connection object to pass the results back

        memory : int | None = None
            The memory in bytes to allocate

        cpu_time : int | None = None
            The cpu time in seconds to allocate

        wall_time : int | None = None
            The amount of total wall time in seconds to limit to

        warnings : bool = True
            Whether to emit pynisher warnings or not.

        wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = False
            Whether to wrap exceptions or not

            Please see `pynisher.__init__` for details

        terminate_child_processes: bool = True
            Whether to clean up all child processes upon completion
        """
        self.func = func
        self.output = output
        self.memory = memory
        self.cpu_time = cpu_time
        self.wall_time = wall_time
        self.warnings = warnings
        self.wrap_errors = wrap_errors
        self.terminate_child_processes = terminate_child_processes

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        """Set process limits and then call the function with the given arguments.

        Note
        ----
        This should usually not be called by the end user, it's main use is in
        Pynisher::run().

        Parameters
        ----------
        *args, **kwargs
            Arguments to the function

        Returns
        -------
        None
        """
        try:
            if self.cpu_time is not None:
                self.limit_cpu_time(self.cpu_time)

            if self.memory is not None:
                # We should probably warn if we exceed the memory usage before
                # any limitation is set
                memusage = Monitor().memory("B")
                if self.memory <= memusage:
                    self._raise_warning(
                        f"Current memory usage in new process is {memusage}B but "
                        f" setting limit to {self.memory}B. Likely to fail, try"
                        f" increasing the memory limit"
                    )
                self.limit_memory(self.memory)

            # Call our function
            result = self.func(*args, **kwargs)
            error = None
            tb = None

        except Exception as e:
            result = None
            error = e
            tb = "".join(traceback.format_exception(*sys.exc_info()))

        if error is not None:
            error = self._wrap_error(error, *args, **kwargs)

        # Now let's try to send the result back
        response = (result, error, tb)
        try:
            self.output.send(response)
            self.output.close()
            if self.terminate_child_processes is True:
                terminate_process(timeout=2, children=True, parent=False)
            return
        except Exception as e:
            failed_send_tb = "".join(traceback.format_exception(*sys.exc_info()))
            failed_send_response = (None, e, failed_send_tb)

        # We failed to send the result back, lets try to send the errors we got from
        # doing so
        try:
            self.output.send(failed_send_response)
            self.output.close()
            if self.terminate_child_processes is True:
                terminate_process(timeout=2, children=True, parent=False)
            return
        except Exception:
            pass

        # One last effot to send the smallest None through
        last_response_attempt = None
        try:
            self.output.send(last_response_attempt)
            self.output.close()
            if self.terminate_child_processes is True:
                terminate_process(timeout=2, children=True, parent=False)
            return
        except Exception:
            pass

        # Look, everything failed, try kill any child processes and just kill this
        # process is we can't for memory reasons
        self.output.close()
        try:
            if self.terminate_child_processes is True:
                terminate_process(timeout=2, children=True, parent=False)
        except MemoryError:
            os.kill(os.getpid(), signal.SIGSEGV)
        finally:
            return

    @staticmethod
    def create(
        func: Callable,
        output: Connection,
        memory: int | None = None,
        cpu_time: int | None = None,
        wall_time: int | None = None,
        warnings: bool = True,
        wrap_errors: bool | list[str | Type[Exception]] | dict[str, Any] = False,
        terminate_child_processes: bool = True,
    ) -> Limiter:
        """For full documentation, see __init__."""
        # NOTE: __init__ param duplication
        #
        #   I'm not delighted by the duplication of the __init__ params but this keeps
        #   things typesafe and referencable in case we need to validate
        arguments = {
            "func": func,
            "output": output,
            "memory": memory,
            "cpu_time": cpu_time,
            "warnings": warnings,
            "wrap_errors": wrap_errors,
            "terminate_child_processes": terminate_child_processes,
        }

        # There is probably a lot more identifiers but for now this covers our use case
        system_name = platform.system().lower()

        # NOTE: Imports inside if statements
        #
        #   This serves two purposes, one is to prevent cyclical imports, as they need
        #   to inherit Limiter from this file, while this file needs to import them too,
        #   creating a circular dependancy... unless they're imported later.
        #
        #   Secondly, different systems will have different modules available.
        #   For example, the `resources` module is not avialable on Windows and so
        #   importing that would cause issues.
        #
        if system_name.startswith("linux"):
            from pynisher.limiters.linux import LimiterLinux

            return LimiterLinux(**arguments)  # type: ignore

        elif system_name.startswith("darwin"):
            from pynisher.limiters.mac import LimiterMac

            return LimiterMac(**arguments)  # type: ignore

        elif system_name.startswith("win"):
            from pynisher.limiters.windows import LimiterWindows

            return LimiterWindows(**arguments)  # type: ignore

        else:
            raise NotImplementedError(
                f"We currently don't support your system: {platform}"
            )

    @abstractmethod
    def limit_memory(self, memory: int) -> None:
        """Limit's the memory of this process."""
        ...

    @abstractmethod
    def limit_cpu_time(self, cpu_time: int) -> None:
        """Limit's the cpu time of this process."""
        ...

    def _raise_warning(self, msg: str) -> None:
        if self.warnings is True:
            print(msg, file=sys.stderr)

    def _wrap_error(self, err: Exception, *args: Any, **kwargs: Any) -> Exception:
        _wrap_message = f"Wrapped Exception {type(err).__name__} - {err}"

        # Catch memory errors first, these don't count as `wrap_errors=False`
        # as we need to catch memory errors that occur due to limits
        if self.memory and (
            isinstance(err, MemoryError)
            or (
                sys.platform.lower().startswith("win")
                and isinstance(err, OSError)
                and getattr(err, "winerr", None) == WIN_ERROR_COMMITMENT_LIMIT
            )
        ):
            err = MemoryLimitException(
                f"Not enough memory to run function ({self.memory}B)."
                f"\n{callstring(self.func, *args, **kwargs)}"
            )

        if self.cpu_time and isinstance(err, CpuTimeoutException):
            err = CpuTimeoutException(
                f"Did not finish in cpu time ({self.cpu_time}s)"
                f"\n{callstring(self.func, *args, **kwargs)}"
            )

        if self.wrap_errors is False:
            return err

        if self.wrap_errors is True:
            return PynisherException(_wrap_message)

        if isinstance(self.wrap_errors, (list, set, tuple)):
            mapping = {"pynisher": self.wrap_errors}
        elif isinstance(self.wrap_errors, dict):
            mapping = self.wrap_errors
        else:
            raise ValueError(f"Arg `wrap_errors` is ill formatted {self.wrap_errors}")

        if self.cpu_time is not None and "cpu_time" in mapping:
            if any(is_err(err, err_type) for err_type in mapping["cpu_time"]):
                return CpuTimeoutException(_wrap_message)

        if self.wall_time is not None and "wall_time" in mapping:
            if any(is_err(err, err_type) for err_type in mapping["wall_time"]):
                return WallTimeoutException(_wrap_message)

        if self.memory is not None and "memory" in mapping:
            for t in mapping["memory"]:
                # Windows specific errors
                if isinstance(t, tuple) and len(t) == 3:
                    errT, errno, winerr = t

                    is_type = is_err(err, errT)
                    has_errno = getattr(err, "errno", None) == errno
                    has_winerr = getattr(err, "winerr", None) == winerr

                    if is_type and has_errno and has_winerr:
                        return MemoryLimitException(_wrap_message)

                # OSError with codes
                elif isinstance(t, tuple) and len(t) == 2:
                    errT, errno, winerr = t

                    is_type = is_err(err, errT)
                    has_errno = getattr(err, "errno", None) == errno

                    if is_type and has_errno:
                        return MemoryLimitException(_wrap_message)

                # `t` is just a straight Exception
                elif is_err(err, t):  # type: ignore
                    return MemoryLimitException(_wrap_message)

        if "pynisher" in mapping:
            if any(is_err(err, err_type) for err_type in mapping["pynisher"]):
                return PynisherException(_wrap_message)

        return err



from __future__ import annotations

import platform
import sys
from functools import lru_cache

if sys.platform.lower().startswith("win"):
    contexts = ["spawn"]
else:
    contexts = ["fork", "spawn", "forkserver"]


@lru_cache(maxsize=None)
def _has_pywin32() -> bool:
    """Check if system has pywin32 accessible"""
    try:
        import win32api  # noqa

        return True
    except Exception:
        return False


def supports(limit: str) -> bool:
    """Check if pynisher supports a given feature

    Parameters
    ----------
    limit: "wall_time" | "cpu_time" | "memory" | "decorator"
        The kind of feature to check support for

    Returns
    -------
    bool
        Whether it is supported or not
    """
    mapping = {
        "wall_time": supports_walltime,
        "cpu_time": supports_cputime,
        "memory": supports_memory,
        "decorator": supports_limit_decorator,
    }
    func = mapping.get(limit.lower(), None)
    if func is not None:
        return func()
    else:
        raise ValueError(f"Not a known feature, must be one of {list(mapping.keys())}")


def supports_walltime() -> bool:
    """Check if wall time is supported on this system

    Check respective "pynisher/limiters/<platform>.py"

    Returns
    -------
    bool
        Whether it's supported or not
    """
    plat = sys.platform.lower()
    if plat.startswith("linux"):
        return True
    elif plat.startswith("darwin"):
        return True
    elif plat.startswith("win"):
        return True
    else:
        raise NotImplementedError(f"Unknown system {platform.platform()}")


def supports_cputime() -> bool:
    """Check if cpu time is supported on this system

    Returns
    -------
    bool
        Whether it's supported or not
    """
    plat = sys.platform.lower()
    if plat.startswith("linux"):
        return True
    elif plat.startswith("darwin"):
        return True
    elif plat.startswith("win"):
        return _has_pywin32()
    else:
        raise NotImplementedError(f"Unknown system {platform.platform()}")


def supports_memory() -> bool:
    """Check if memory limit is supported on this system

    * Linux - Yes
    * Darwin - No
    * Windows - If `pywin32` installed correctly

    Check respective "pynisher/limiters/<platform>.py"

    Returns
    -------
    bool
        Whether it's supported or not
    """
    plat = sys.platform.lower()
    if plat.startswith("linux"):
        return True
    elif plat.startswith("darwin"):
        return False
    elif plat.startswith("win"):
        return _has_pywin32()
    else:
        raise NotImplementedError(f"Unknown system {platform.platform()}")


def supports_limit_decorator() -> bool:
    """Whether using the decorator @restricted is supported

    Check `pynisher::restricted` for why

    Returns
    -------
    bool
        Whether using @restricted is supported
    """
    plat = sys.platform.lower()
    if plat.startswith("linux"):
        return True
    elif plat.startswith("darwin"):
        return False
    elif plat.startswith("win"):
        return False
    else:
        raise NotImplementedError(f"Unknown system {platform.platform()}")



from __future__ import annotations

from typing import Any

import sys

from pynisher.limiters.limiter import Limiter

_win32_import_error_msg = """
Couldn't limit `memory` as `pywin32` failed to import.

If using a Conda environment with Python 3.8 or 3.9 and you installed `pynisher`
or a package relying on `pynisher` with `pip` you must install `pywin32`
with `conda` manually:

* `conda install pywin32`.

Please see this issue for more:
* https://github.com/mhammond/pywin32/issues/1865

pywin32
* https://github.com/mhammond/pywin32
"""

# Not sure it's actual type
WIN32JOB = Any


class LimiterWindows(Limiter):
    def job(self) -> WIN32JOB:
        """Get the job associated with this process.

        Caches between calls for safety.

        Returns
        -------
        WIN32JOB
            A Windows job which consists of one or more processes. In this case
            we just have the one process
        """
        if not hasattr(self, "_job"):
            # First try to import stuff, an easy exception to catch and give good
            # information about
            try:
                import win32api
                import win32job
                import winerror
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(_win32_import_error_msg) from e

            # Here, we assign it to a windows "Job", whatever that is
            # If the process is already assigned to a job, then we have
            # to check if it's less than Windows 8 because apparently
            # nested jobs aren't supported there
            try:
                job = win32job.CreateJobObject(None, "")

                process = win32api.GetCurrentProcess()
                win32job.AssignProcessToJobObject(job, process)

                self._job = job

            except win32job.error as e:
                if (
                    e.winerror != winerror.ERROR_ACCESS_DENIED
                    or sys.getwindowsversion() >= (6, 2)  # type: ignore
                    or not win32job.IsProcessInJob(process, None)
                ):
                    raise e

                msg = (
                    "The process is already in a job."
                    " Nested jobs are not supported prior to Windows 8."
                )
                raise RuntimeError(msg) from e

        return self._job

    def limit_memory(self, memory: int) -> None:
        """Limit's the memory of this process."""
        # https://stackoverflow.com/a/54958772/5332072

        # First try to import stuff, an easy exception to catch and give good
        # information about
        try:
            import win32job
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(_win32_import_error_msg) from e

        job = self.job()

        # Get the information for the job object we created
        enum_for_info = win32job.JobObjectExtendedLimitInformation
        info = win32job.QueryInformationJobObject(job, enum_for_info)

        # This appears to by bytes by looking for "JOB_OBJECT_LIMIT_JOB_MEMORY"
        # on github, specifically Python code
        info["ProcessMemoryLimit"] = memory

        # Mask of which limit flag to set
        mask_limit_memory = win32job.JOB_OBJECT_LIMIT_PROCESS_MEMORY
        info["BasicLimitInformation"]["LimitFlags"] |= mask_limit_memory

        # Finally set the new information
        win32job.SetInformationJobObject(job, enum_for_info, info)

    def limit_cpu_time(self, cpu_time: int) -> None:
        """Limit's the cpu time of this process.

        Note
        ----
        Unfortunatly, when windows terminates a process, it doesn't use signaling
        methods like unix based systems, it just kills it and gives no chance for clean
        up. Pynisher detects this in the parent process by looking for an empty response
        from the pipe, checking that the process was killed non-gracefully, and then
        seeing if cpu_time was set.

        Parameters
        ----------
        cpu_time: int
            How many seconds to limit the process for
        """
        # First try to import stuff, an easy exception to catch and give good
        # information about
        try:
            import win32job
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(_win32_import_error_msg) from e

        job = self.job()

        # Get the information for the job object we created
        enum_for_info = win32job.JobObjectBasicLimitInformation
        info = win32job.QueryInformationJobObject(job, enum_for_info)

        # Set the time limit
        time = round(cpu_time * 10_000_000)  # In 100ns units (1e+9 / 100)
        info["PerProcessUserTimeLimit"] = time

        # Activate the flag to turn on the limiting of cput time
        flag = win32job.JOB_OBJECT_LIMIT_PROCESS_TIME
        info["LimitFlags"] |= flag

        # Finally set the new information
        win32job.SetInformationJobObject(job, enum_for_info, info)



from pathlib import Path

from setuptools import find_packages, setup

readme = Path(__file__).parent / "README.md"

with readme.open("r") as fh:
    long_description = fh.read()

setup(
    name="pynisher",
    version="1.0.6",
    packages=find_packages(where=".", include=["pynisher*"], exclude=["test*"]),
    include_package_data=True,
    install_requires=[
        "psutil",
        "typing_extensions",
        "pywin32; platform_system=='Windows'",
    ],
    extras_require={
        "test": [
            "pytest",
            "pre-commit",
            "pytest-cov",
            "pytest-forked",
            "pydocstyle[toml]",
            "isort",
            "black",
            "flake8",
            "mypy",
            "scikit-learn",
        ],
    },
    author=(
        "Stefan Falkner, Christina Hernandez-Wunsch, Samuel Mueller,"
        "Matthias Feurer, Francisco Rivera, Eddie Bergman and Rene Sass",
    ),
    author_email="feurerm@informatik.uni-freiburg.de",
    description="A library to limit the resources used by functions using subprocesses",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="resources",
    license="MIT",
    url="https://github.com/automl/pynisher",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.7",
)
